/*
 * XLife Copyright 1989 Jon Bennett jb7m+@andrew.cmu.edu, jcrb@cs.cmu.edu
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * A lot of modifications were added at 2001, 2011-13 by Vladimir Lidovski vol.litwr@gmail.com
 * (C) This version of XLife may be used under the same conditions as mentioned above
 * $Id: tile.c 107 2013-01-30 15:05:05Z micro $
 */

/*
 * This module encapsulates cell-box handling. Everything else in the program
 * except possibly the evolution function should go through these functions
 * to change the board state. Entry points are:
 *
 * void initcells(context)        -- initialize active pattern
 *
 * int lookcell(context, x, y)    -- get state of cell
 *
 * void chgcell(context, x, y, c) -- set cell to state c
 *
 * void changesize(n)             -- change the cellbox allocation size
 *
 * void center()                  -- center viewport on barycenter of pattern
 *
 * void clear()                   -- clear the universe, freeing all storage
 *
 * void redisplay(int showgrid)   -- update visible display of active board
 *
 * void redrawscreen(int showgrid)-- redraw, assuming nothing about old state
 *
 * void saveall()                 -- save entire board state to file
 *
 * This code knows that cells are packed in 8x8 tiles on a hash list, but
 * it doesn't know anything about the internals of cell representation within
 * cell boxes.  The function changesize() changes the size of cellboxes.
 *
 * This code relies on the existence of a triad of functions:
 *
 * getcell(ptr, dx, dy)      -- get state of cell at x, y in box *ptr
 * setcell(ptr, dx, dy, val) -- set cell at x, y in box *ptr to state val
 *
 * to access cell state. It also relies on the existence of these functions:
 *
 * displaybox(x, y, ptr)     -- post box state to Xrect arrays
 * fetchtile(context, x, y)  -- get tile containing cell (x, y), NULL if none
 * maketile(context, x, y)   -- get tile holding cell (x, y), creating if none
 */

#include <stddef.h>	/* we need offsetof() */
#include <stdlib.h>
#include "defs.h"
#include "map.h"
#include "patchlevel.h"
#include "colors.h"
#include "tile.h"
#include "file.h"
#include "framebuffer.h"

#define MARK		'*'
#define SPACE		'.'
#define	USL_MAX		4294967295U	/* max decimal value of "unsigned" */

#define XMASK 0xfffffff8
#define YMASK 0xfffffff8

/* public variables */
pattern active, tentative;
tile *freep, *curtile;
static unsigned hxm = ((1 << HASHBITS - HASHBITS/2) - 1) << 3,
   hym = (1 << HASHBITS/2) - 1, pxs;
static int hxs = HASHBITS/2 - 3;
static XTextProperty windowName;
static char window_name_pattern[61], *window_name = window_name_pattern;

//#define HASH(x,y) (shl((x & HXM),HXS) + ((y >> 3) & HYM))

unsigned HASH(register unsigned x, register unsigned y) {
    return shl((x & hxm),hxs) | ((y >> 3) & hym);
}

void hashstat(void) {
   unsigned hashsum = 0, hashcount = 0, i;
   tile *ptr;
   for (i = 0; i < HASHSIZE; i++)
      if (ptr = active.hashlist[i]) {
          do hashsum++; while (ptr = ptr->hnext);
          hashcount++;
      }
   sprintf(inpbuf, "Hash: X=%u Y=%u [%u] filled=%.4f badness=%.3f",
	     HASHBITS - 3 - hxs, 3 + hxs, HASHBITS, (double)hashcount/HASHSIZE,
	                      (hashcount ? (double)hashsum/hashcount : 1) - 1);
}

void fixhash(pattern *p) {
   tile *ptr;
   memset(p->hashlist, 0, HASHSIZE*sizeof(tile*));
   for (ptr = p->tiles; ptr; ptr = ptr->next) {
      unsigned hv = HASH(ptr->x, ptr->y);
      if (ptr->hnext = p->hashlist[hv])
             ptr->hnext->hfore = ptr;
      p->hashlist[hv] = ptr;
   }
}

void adjhash(pattern *p) {
   unsigned xc = 0, yc = 0;
   if (p->tilecount < 2048) return;
   bounding_box(p);
   while (p->xmax - p->xmin >> 3 + xc++);
   while (p->ymax - p->ymin >> 3 + yc++);
   xc = HASHBITS/(1 + (double)yc/xc);
   if (xc < 4)
       xc = 4;
   else if (xc > HASHBITS - 4)
       xc = HASHBITS - 4;
   if (xc == pxs) return;
   yc = HASHBITS - (pxs = xc);
   hxm = ((1 << xc) - 1) << 3,
   hym = (1 << yc) - 1;
   hxs = yc - 3;
   fixhash(p);
}

void setscale(const int sc) { /* preset scale for X rectangles describing active board */
    int min = 1, max = 1;
    if (sc > 0) {
	int	i, j, size = (1 << sc) - (sc > 1);

	for (i = 0; i < MAXSTATES; i++)
	    for (j = 0; j < MAXCHANGE; j++) {
		rects[i][j].height = size;
		rects[i][j].width = size;
	    }
    }
    if (sc < 0)
       min = 1 << -sc;
    else
       max = 1 << sc;
    sprintf(window_name_pattern, "XLife v%d.%d.%d: a cellular-automaton laboratory [%d:%d]",
       majorversion, minorversion, patchlevel, min, max);
    if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
       fprintf(stderr, "structure allocation for WindowName failed.\n");
       exit(5);
    }
    XSetWMName(disp, mainw, &windowName);
}

void initcells(pattern *context) { /* initialize the cell hash list */
    memset(context, '\0', sizeof(pattern));
}

void chgcell(pattern *context, coord_t x, coord_t y, cell_t c) {
/* turn a cell to state ON */
    tile *ptr = maketile(context, x, y);
    unsigned ydx = y - ptr->y, xdx = x - ptr->x;
    if (c) ptr->dead = 0;
    setcell(&ptr->cells, xdx, ydx, c);
}

void changesize(int newsize) { /* change the tile allocation size */
    if (tilesize != offsetof(tile, cells) + newsize) {
	tile	*ptr, *nptr;
	clear_pattern(&active);
	for (ptr = freep; ptr; ptr = nptr) {
	    nptr = ptr->next;
	    free(ptr);
	}
	freep = 0;
	tilesize = offsetof(tile, cells) + newsize;
        DoExpose(lifew);
    }
}

void killtile(pattern *context, tile *ptr) {
/* nuke a tile, reclaiming all associated storage and fixing neighbor links */
    unsigned hv = HASH(ptr->x, ptr->y);
    if (ptr != context->tiles)
	ptr->fore->next = ptr->next;
    else
	context->tiles = ptr->next;
    if (ptr == context->hashlist[hv])
	context->hashlist[hv] = ptr->hnext;
    else
	ptr->hfore->hnext = ptr->hnext;
    if (ptr->next) ptr->next->fore = ptr->fore;
    if (ptr->hnext) ptr->hnext->hfore = ptr->hfore;
    if (ptr->rt) ptr->rt->lf = 0;
    if (ptr->lf) ptr->lf->rt = 0;
    if (ptr->up) ptr->up->dn = 0;
    if (ptr->dn) ptr->dn->up = 0;
    ptr->next = freep;
    freep = ptr;
    context->tilecount--;
}

tile *createtile(pattern *context, coord_t x, coord_t y, unsigned hv) {
/* create a tile to hold cells near life-world coordinates x, y */
/* universe it belongs in */
/* associated life-world coordinates */
/* tile-retrieval hash value */
    tile *ptr;

    if (freep == NULL) {
	if ((ptr = (tile*)malloc(tilesize)) == NULL){
	    perror("create: malloc error: ");
	    exit(-1);
	}
    }
    else {
	ptr = freep;
	freep = freep->next;
    }
    memset(ptr, '\0', tilesize);
    ptr->next = context->tiles;
    context->tiles = ptr;

    if (ptr->next)
	ptr->next->fore = ptr;
    ptr->hnext = context->hashlist[hv];
    context->hashlist[hv] = ptr;
    if (ptr->hnext)
	ptr->hnext->hfore = ptr;
    ptr->x = x;
    ptr->y = y;
    context->tilecount++;
    return ptr;
}


static tile *findtile(pattern *context, coord_t x, coord_t y, unsigned hv) {
    tile *ptr = context->hashlist[hv];

    while (ptr) {
	    if (ptr->x == x && ptr->y == y)
		return ptr;
	    ptr = ptr->hnext;
    }
    return 0;
}

int lookcell(pattern *context, coord_t x, coord_t y) {
/* get the state of a cell from its world coordinates */
    tile *ptr;
    unsigned hv;
    coord_t hx, hy;

    hx = x & XMASK;
    hy = y & YMASK;
    hv = HASH(x, y);
    if (ptr = findtile(context, hx, hy, hv))
	return getcell(&ptr->cells, x - hx, y - hy);
    else
	return 0;
}

#define checktore() if (topology == 'T' && context == &active) {\
         if (x >= x_max_limit)\
            x = x - x_max_limit + x_min_limit;\
         else if (x < x_min_limit)\
            x = x - x_min_limit + x_max_limit;\
         if (y >= y_max_limit)\
            y = y - y_max_limit + y_min_limit;\
         else if (y < y_min_limit)\
            y = y - y_min_limit + y_max_limit;\
    }

tile *fetchtile(pattern *context, coord_t x, coord_t y) {
    checktore();
    return findtile(context, x, y, HASH(x, y));
}

tile *maketile(pattern *context, coord_t x, coord_t y) {
    tile *ptr;
    unsigned hv;
    checktore();
    hv = HASH(x, y);
    x &= XMASK;
    y &= YMASK;
    if (ptr = findtile(context, x, y, hv))
       return ptr;
    else
       return createtile(context, x, y, hv);
}

void center(void) {
/* center the display on the `center of mass' of the live boxes */
    double x = 0, y = 0;
    int ctr = 0;
    tile *ptr;
    for (ptr = active.tiles; ptr; ptr = ptr->next)
	if (!ptr->dead){
	    x += ptr->x;
	    y += ptr->y;
	    ctr++;
	}
    if (ctr > 0) {
       x /= ctr;
       y /= ctr;
    }
    else
      x = xorigin, y = yorigin;
    xpos = x - shr(width/2, scale);
    ypos = y - shr(height/2, scale);
    display_move();
}

int set_torus(coord_t x, coord_t y) {
    x = (x + 7 >> 3)*8;
    y = (y + 7 >> 3)*8;
    limits = x_min_limit = x_max_limit = y_min_limit = y_max_limit = 0;
    if (x) {
          x_min_limit = STARTX - (x - x%16)/2;
          x_max_limit = STARTX + (x + x%16)/2;
          limits = 1;
    }
    if (y) {
          y_min_limit = STARTY - (y - y%16)/2;
          y_max_limit = STARTY + (y + y%16)/2;
          limits |= 2;
    }
    return limits;
}

void clr_tile_links() {
   tile *ptr = active.tiles;

   for (; ptr; ptr = ptr->next)
      ptr->up = ptr->dn = ptr->lf = ptr-> rt = 0;
}

void display_move() {
   if (tentative.tiles)
      DoExpose(lifew);
   else
      redisplay(0);
}

static int cmpnum(const void *e1, const void *e2) {
  if (*(coord_t*)e1 < *(coord_t*)e2)
    return -1;
  return *(coord_t*)e1 > *(coord_t*)e2;
}

void median() {
/* return the median coordinates of the live cells in the context */
    coord_t *coordxlist, *coordylist;
    int ctr = 0, x = xorigin, y = yorigin;
    tile *ptr;

    if (!(coordxlist = (coord_t*)malloc(sizeof(coord_t)*active.tilecount)))
	return;
    if (!(coordylist = (coord_t*)malloc(sizeof(coord_t)*active.tilecount))) {
        free(coordxlist);
	return;
    }
    for (ptr = active.tiles; ptr; ptr = ptr->next){
	if(!ptr->dead) {
	    coordxlist[ctr] = ptr->x;
	    coordylist[ctr++] = ptr->y;
	}
    }
    if (ctr > 0) {
	qsort(coordxlist, ctr, sizeof(coord_t), cmpnum);
	qsort(coordylist, ctr, sizeof(coord_t), cmpnum);
	x = coordxlist[ctr/2];
	y = coordylist[ctr/2];
    }
    free(coordylist);
    free(coordxlist);
    xpos = x - shr(width/2, scale);
    ypos = y - shr(height/2, scale);
    display_move();
}

void current_tile() {
    if (curtile == NULL)
       curtile = active.tiles;
    if (curtile) {
       coord_t xnew, ynew;
       int lflag = 0;
       xnew = curtile->x - shr(width/2, scale);
       ynew = curtile->y - shr(height/2, scale);
       while (curtile->dead || labs((long)xnew - (long)xpos) < 10 << 6 - scale && labs((long)ynew - (long)ypos) < 10 << 6 - scale) {
          curtile = curtile->next;
          if (curtile == NULL) {
             if (lflag) return;
             curtile = active.tiles;
             lflag++;
          }
          xnew = curtile->x - shr(width/2, scale);
          ynew = curtile->y - shr(height/2, scale);
       }
       xpos = xnew;
       ypos = ynew;
       curtile = curtile->next;
       DoExpose(lifew);
    }
}

cellcount_t cellcnt(pattern *pp) {
   int i;
   cellcount_t sum = 0;

   for (i = 1; i < maxstates; i++)
      sum += pp->cellcount[i];
   return sum;
}

void displaystats(void) {
    sprintf(inpbuf, "Generations: %lu", active.generations);
    if (cellmass) strcat(inpbuf, "*");
    sprintf(inpbuf + strlen(inpbuf), ", Tiles: %u", active.tilecount);
    if (dispboxes)
	sprintf(inpbuf + strlen(inpbuf), ", Cells: %u", cellcnt(&active));
    if (dispchanges)
        sprintf(inpbuf + strlen(inpbuf), ", Changes: %u", active.chgcount);
    DoExpose(inputw);
}

void clear_pattern(pattern *pp) {
/* free given pattern */
    tile *ptr, *nptr;
    initcells(pp);
    ptr = pp->tiles;
    while (ptr) {
	nptr = ptr->next;
	free(ptr);
	ptr = nptr;
    }
    pp->tiles = NULL;
    pp->generations = 0;
    pp->tilecount = 0;
    memset(pp->cellcount, 0, sizeof(cellcount_t)*maxstates);
    fb_clear();
}

void clear(void) {
/* clear the board or tentative pattern, freeing all storage */
    XClearWindow(disp, inputw);
    if (tentative.tiles) {
	clear_pattern(&tentative);
	free_loadscript();    		/* restart load script */
    }
    else
    {
        map_init();
	clear_pattern(&active);
	state = STOP;
	displaystats();
    }
    DoExpose(lifew);
}

void clear_all(void) {
/* clear the board or tentative pattern, freeing all storage */
    XClearWindow(disp, inputw);
    if (tentative.tiles) {
        clear_pattern(&tentative);
        free_loadscript();              /* restart load script */
    }
    clear_pattern(&active);
    /*state = STOP;*/
    displaystats();
    DoExpose(lifew);
}

#define VFREQ 50
#define SYNCN 1
static int fb_sync() {
#if SYNCN == 1
   return 1;
#else
   static struct timeval prev, cur;
   gettimeofday(&cur, 0);
   if ((cur.tv_sec - prev.tv_sec)*1000000 + cur.tv_usec - prev.tv_usec > 1000000/VFREQ) {
      prev = cur;
      return 1;
   }
   return state == STOP || force_fb;
#endif
}

static int onscreen(coord_t x, coord_t y) {
   /*this is a bit complicated due to the big torus calculations */
   int f = (scale < -3 ? 1 << -3 - scale : 1),
      px_min = x > xpos - BOXSIZE*f, py_min = y > ypos - BOXSIZE*f,
      px_max = x < xpos + BOXSIZE*f + shr(width, scale),
      py_max = y < ypos + BOXSIZE*f + shr(height, scale);
   return (px_min && px_max
     || (px_min || px_max) && (xpos - BOXSIZE*f > xpos + BOXSIZE*f + shr(width, scale)))
     && (py_min && py_max
     || (py_min || py_max) && (ypos - BOXSIZE*f > ypos + BOXSIZE*f + shr(height, scale)));
}

void redisplay(int showgrid) {
/* re-display the visible part of the board */
    tile *ptr;
    coord_t x, y, tentx, tenty;
    int xmaxstates = maxstates + 2*(pseudocolor && maxstates == 2), curstate, q = 0;

    if (state == HIDE) return;
    if (dispmesh && scale > 1 && showgrid) {
       coord_t cx = width >> scale, cy = height >> scale;
       int small = 1 << scale;

       for (x = 0; x <= cx; x++)
            if (x%5 == 0)
               XDrawLine(disp, lifew, cellgc[GRID_ECOLOR], x*small - 1, 0, x*small - 1, height);
            else
               XDrawLine(disp, lifew, cellgc[GRID_COLOR], x*small - 1, 0, x*small - 1, height);
       for (y = 0; y <= cy; y++) {
            if (y%5 == 0)
               XDrawLine(disp, lifew, cellgc[GRID_ECOLOR], 0, y*small - 1, width, y*small - 1);
            else
               XDrawLine(disp, lifew, cellgc[GRID_COLOR], 0, y*small - 1, width, y*small - 1);
        }
    }
    for (curstate = 0; curstate < xmaxstates; curstate++)
	chgpoints[curstate] = chgrects[curstate] = 0;
    for (ptr = active.tiles; ptr; ptr = ptr->next) {
	x = ptr->x;
	y = ptr->y;
	if (onscreen(x, y)) {
            if (scale < -3) {
               tentx = (x >> -scale) << -scale;
               tenty = (y >> -scale) << -scale;
               if (x != tentx || y != tenty) {
                  if (!fetchtile(&active, tentx, tenty)) {
                     maketile(&active, tentx, tenty);
                     q++;
                  }
                  continue;
               }
            }
	    displaybox(x, y, &ptr->cells);
	}
    }
    if (q)
        for (ptr = active.tiles; q--; ptr = ptr->next)
           displaybox(ptr->x, ptr->y, &ptr->cells);
    if (fb_sync()) {
       fb_flush();
       if (scale <= 0) {
                for (curstate = 0; curstate < xmaxstates; curstate++)
                    if (chgpoints[curstate]) {
                        XDrawPoints(disp, lifew, cellgc[curstate], points[curstate],
                                      chgpoints[curstate], CoordModeOrigin);
                        chgpoints[curstate] = 0;
                    }
            }
            else
                for (curstate = 0; curstate < xmaxstates; curstate++)
                    if (chgrects[curstate]) {
                        XFillRectangles(disp, lifew, cellgc[curstate],
                                        rects[curstate], chgrects[curstate]);
                        chgrects[curstate] = 0;
                    }
       /* draw pivot of tentative points and box them */
       if (tentative.tiles) {
          if (!wireframe) boxpattern(WHITE_EVER); /* white outline box */
          /* draw tentative points with appropriate transformation */
          for (ptr = tentative.tiles; ptr; ptr = ptr->next) {
              tentx = ptr->x - STARTX;
              tenty = ptr->y - STARTY;
              if (onscreen(tx(tentx,tenty,txx,txy) + loadx,
                           ty(tentx,tenty,tyx,tyy) + loady)) {
                  if (scale < -3) {
                     tentx = (tentx >> -scale) << -scale;
                     tenty = (tenty >> -scale) << -scale;
                  }
                  trdisplaybox(tentx, tenty, &ptr->cells);
                  if (scale <= 0) {
                      for (curstate = 0; curstate < maxstates; curstate++)
                          if (chgpoints[curstate]) {
                              XDrawPoints(disp, lifew, cellgc[curstate],
                                          points[curstate], chgpoints[curstate],
                                          CoordModeOrigin);
                              chgpoints[curstate] = 0;
                          }
                  }
                  else
                      for (curstate = 0; curstate < maxstates; curstate++)
                          if (chgrects[curstate]) {
                              if (wireframe)
                                   XDrawRectangles(disp, lifew, cellgc[LOAD_BOX],
                                              rects[curstate], chgrects[curstate]);
                              else
                                   XFillRectangles(disp, lifew, cellgc[curstate],
                                              rects[curstate], chgrects[curstate]);
                              chgrects[curstate] = 0;
                          }
              }
          }
	  drawpivot(shl(loadx - xpos, scale), shl(loady - ypos, scale));
       }
       XFlush(disp);
    }
}

#define conf_y(y1, y2) if (ypos <= y_max_limit && shr(height, scale) >= (int)(y_min_limit - ypos)) {\
              if (ypos < y_min_limit) {\
                 y1 = shl(y_min_limit - ypos, scale);\
                 if (shr(height, scale) > (int)(y_max_limit - ypos))\
                   y2 = shl(y_max_limit - y_min_limit, scale);\
                 else\
                   y2 = height - shl(y_min_limit - ypos, scale);\
              }\
              else {\
                 y1 = 0;\
                 if (shr(height, scale) > (int)(y_max_limit - ypos))\
                   y2 = shl(y_max_limit - ypos, scale);\
                 else\
                   y2 = height;\
              }\
              XFillRectangle(disp, lifew, cellgc[GRID_COLOR], x1, y1, x2, y2);\
            }

void redrawscreen(int showgrid) { /* force redraw of entire board */
   if (limits & 2) {
      int x1 = 0, y1, x2, y2;
      if (ypos < y_min_limit)
         XFillRectangle(disp, lifew, cellgc[GRID_COLOR], 0, 0, width,
                                  min(height, shl(y_min_limit - ypos, scale)));
      if (shr(height, scale) > (int)(y_max_limit - ypos))
         if (y_max_limit > ypos)
            XFillRectangle(disp, lifew, cellgc[GRID_COLOR], 0,
                                         shl(y_max_limit - ypos, scale), width,
                                      height - shl(y_max_limit - ypos, scale));
         else
            XFillRectangle(disp, lifew, cellgc[GRID_COLOR], 0, 0, width, height);
      if (limits & 1) {
         if (xpos < x_min_limit
                              && shr(height, scale) > (int)(y_min_limit - ypos)
                                                       && ypos < y_max_limit) {
            if (shr(width, scale) < x_min_limit - xpos)
               x2 = width;
            else
               x2 = shl(x_min_limit - xpos, scale);
            conf_y(y1, y2);
         }
         if (shr(width, scale) > (int)(x_max_limit - xpos)
                              && shr(height, scale) > (int)(y_min_limit - ypos)
                                                       && ypos < y_max_limit) {
            if (xpos > x_max_limit)
               x2 = width;
            else {
               x1 = shl(x_max_limit - xpos, scale);
               x2 = width - shl(x_max_limit - xpos, scale);
            }
            conf_y(y1, y2);
         }
      }
   }
   else if (limits) {
      if (xpos < x_min_limit)
         XFillRectangle(disp, lifew, cellgc[GRID_COLOR], 0, 0,
                           min(width, shl(x_min_limit - xpos, scale)), height);
      if (shr(width, scale) > (int)(x_max_limit - xpos))
         if (x_max_limit > xpos)
            XFillRectangle(disp, lifew, cellgc[GRID_COLOR],
                                             shl(x_max_limit - xpos, scale), 0,
                               width - shl(x_max_limit - xpos, scale), height);
         else
            XFillRectangle(disp, lifew, cellgc[GRID_COLOR], 0, 0, width, height);
   }
   redisplay(showgrid);
}

cellcount_t bounding_box(pattern *context) {
/* get the bounding box of a pattern */
   int dx, dy, state;
   tile *ptr;
   context->ymin = context->xmin = USL_MAX;
   context->ymax = context->xmax = 0;
   if (context->tiles == 0)
      return context->ymin = context->xmin = 0;
   for (state = 1; state < maxstates; state++)
      context->cellcount[state] = 0;
   FOR_CELLS(context, ptr, dx, dy) {
      if (state = getcell(&ptr->cells, dx, dy)) {
         context->cellcount[state]++;
         if (ptr->x + dx < context->xmin)
            context->xmin = ptr->x + dx;
         if (ptr->y + dy < context->ymin)
            context->ymin = ptr->y + dy;
         if (ptr->x + dx > context->xmax)
            context->xmax = ptr->x + dx;
         if (ptr->y + dy > context->ymax)
            context->ymax = ptr->y + dy;
      }
   }
   return cellcnt(context);
}

static void flood_fill(pattern *from, coord_t x, coord_t y, pattern *to) {
/* transplant all points in a blob touching (x,y) between contexts */
    static offsets neighbors[] =
    {
	{-1, -1},
	{-1,  0},
	{-1,  1},
	{ 0, -1},
	{ 0,  1},
	{ 1, -1},
	{ 1,  0},
	{ 1,  1},
#ifdef DEPTH2
	{-2, -2},
	{-2, -1},
	{-2,  0},
	{-2,  1},
	{-2,  2},
	{-1,  2},
	{ 0,  2},
	{ 1,  2},
	{ 2,  2},
	{ 2,  1},
	{ 2,  0},
	{ 2, -1},
	{ 2, -2},
	{ 1, -2},
	{ 0, -2},
	{-1, -2},
#endif /* DEPTH2 */
    };
    int i;

    /* skip empty cells */
    if (!lookcell(from, x, y))
	return;

    /* skip cells that are already colored */
    if (lookcell(to, x, y))
	return;

    /* transfer this cell from the copy to the new context */
    chgcell(to,   x, y, lookcell(from, x, y));
    chgcell(from, x, y, 0);

    for (i = 0; i < sizeof(neighbors)/sizeof(neighbors[0]); i++) {
	int nx = x + neighbors[i].xi;
	int ny = y + neighbors[i].yi;

	/* recursively transfer all its neighbors */
	flood_fill(from, nx, ny, to);
    }
}

static int refcmp(const void *r1, const void *r2) {
/* sort blobs first by least Y coordinate, then by least X coordinate */
    int	cmp = (((pattern*)r1)->ymin - ((pattern*)r2)->ymin);

    if (!cmp)
	cmp = (((pattern*)r1)->xmin - ((pattern*)r2)->xmin);
    return cmp;
}

pattern *analyze(pattern *context) {
/* analyze a pattern into an allocated list of blobs */
    cell_t	state;
    tile	*ptr;
    int		dx, dy, color = 0;
    pattern	*parts, *pp;
    static pattern analysand; /* too big to be automatic */

    initcells(&analysand);

    /* copy the pattern */
    FOR_CELLS(context, ptr, dx, dy)
	if ((state = getcell(&ptr->cells, dx, dy)))
	    chgcell(&analysand, ptr->x + dx, ptr->y + dy, state);
    /*
     * This loop is deceptively simple-looking.  It relies on flood_fill()
     * not returning until it has transpanted the largest blob it can find
     * connected to a cell.
     */
    parts = (pattern*)calloc(sizeof(pattern), 1);
    FOR_CELLS(&analysand, ptr, dx, dy)
	if (getcell(&ptr->cells, dx, dy)) {
	    parts = (pattern *)realloc(parts, sizeof(pattern) * (color + 1));
	    initcells(&parts[color]);
	    flood_fill(&analysand, ptr->x + dx, ptr->y + dy, &parts[color]);
	    color++;
	}

    /* create an extra, blank pattern entry as a fencepost */
    parts = (pattern *)realloc(parts, sizeof(pattern) * (color + 1));
    initcells(&parts[color]);

    /* this cleans up the blob order -- strictly cosmetic */
    for (pp = parts; pp->tiles; pp++)
	bounding_box(pp);
    qsort(parts, color, sizeof(pattern), refcmp);

    /* return the analyzed parts */
    return parts;
}

void saveformatsign(char f, int xoff, int yoff, FILE* ofp) {
   fprintf(ofp, "#%c", f);
   if (xoff || yoff)
      fprintf(ofp, " %d %d", xoff, yoff);
   fputs("\n", ofp);
}

void savesimple(pattern *pp, int xoff, int yoff, FILE *ofp) {
/* save a pattern as a simple picture block */
    coord_t x, y;
    cellcount_t cellcount = bounding_box(pp);
    if (cellcount == 0)
	return;
    saveformatsign('P', xoff, yoff, ofp);
    for (y = pp->ymin; y <= pp->ymax; y++) {
	for (x = pp->xmin; x <= pp->xmax; x++) {
	    int val;
	    if (val = lookcell(pp, x, y)) {
		if (maxstates > 2)
		    fputc(itos(val), ofp);
		else
		    fputc(MARK, ofp);
	    }
	    else
		fputc(SPACE, ofp);
	}
	fputc('\n', ofp);
    }
}

static void savecomplex(pattern *context, FILE *ofp) {
/* do a pattern analysis and save the pattern as a blob sequence */
    pattern	*pp, *parts = analyze(context);
    int		xcenter, ycenter, blobcount = 0;
    if (parts == NULL) {
	perror("Pattern analysis failed; too many blobs in picture.\n");
	return;
    }
    for (pp = parts; pp->tiles; pp++)
	blobcount++;
    if (blobcount > 1)
	fprintf(ofp, "#C %d blobs\n", blobcount);
    for (pp = parts; pp->tiles; pp++)
	savesimple(pp, pp->xmin - STARTX, pp->ymin - STARTY, ofp);
}

char* rlestr(int c) {
   static char s[3];
   if (c < 25) {
      s[1] = 0;
      if (c == 0)
        s[0] = 'b';
      else if (c == 1)
        s[0] = 'o';
      else
        s[0] = c + '@';
   }
   else {
      s[0] = 'o' + (c - 1)/24;
      s[1] = (c - 1)%24 + 'A';
      s[2] = 0;
   }
   return s;
}

#define rleflush(n) if (strlen(buf) > n) {\
       t = *p;\
       *p = '\0';\
       fprintf(ofp, "%s\n", buf);\
       *p = t;\
       strcpy(buf, p);\
   }\
   p = buf + strlen(buf);


void saveall(FILE *ofp, char mode) {
/* save the entire board contents (or the tentative pattern if there is one) */
    tile *ptr;
    char buf[100], *p = buf, t;
    int dx, dy, val;
    coord_t x = 0, y;
    cellcount_t cellcount;
    pattern *pp = tentative.tiles ? &tentative : &active;
    fprintf(ofp, "##XLife\n");
    if (fname[0])
        fprintf(ofp, "#N %s\n", fname);
    else if (stashed[0]) {
        char *p = strrchr(stashed, '/');
        if (p)
             p++;
        else
             p = stashed;
        if (strchr(p, ':'))
           fprintf(ofp, "#N %s\n", p);
    }
    stamp("#0", ofp);
    if (outcome[0])
        fprintf(ofp, "#K %s\n", outcome);
    while (x < numcomments)
	fprintf(ofp, "#C %s\n", comments[x++]);
    if (fname[0])
	fprintf(ofp, "#N %s\n", fname);
    strcpy(buf, active_rules);
    if (!strcmp(buf, "life"))
        strcpy(buf, "23/3");
    t = 'U';
    if (strchr(buf, '$') || !strcmp(buf, "wireworld") || !strcmp(buf, "brian's brain")
             || strchr(buf, '/') || !strcmp(buf, "life+") || !strcmp(buf, "lifehistory"))
        t = 'T';
    fprintf(ofp, "#%c %s", t, buf);
    if (topology != 'Q')
        fprintf(ofp, ":%c%d,%d", topology, x_max_limit - x_min_limit, y_max_limit - y_min_limit);
    fprintf(ofp, "\n");
    if (*colorfile)
       fprintf(ofp, "#X %s\n", strrchr(colorfile, '/') + 1);
    buf[0] = 0;
/*
 * Determine the bounding box of the live cells.
 */
    cellcount = bounding_box(pp);
/* caller may want save to shortest format */
    if (mode == '\0')
	if ((pp->ymax - pp->ymin + 1)*(pp->xmax - pp->xmin + 1) > cellcount * 8)
	    mode = 'R';
	else
	    mode = 'P';

/* now that we have the bounding box, we can gen the other formats */
    switch (mode) {
	case 'A': /* save in absolute mode */
	  fputs("#A\n", ofp);
	  FOR_CELLS(pp, ptr, dx, dy)
			if (val = getcell(&ptr->cells, dx, dy))
			    if (maxstates > 2)
				fprintf(ofp, "%d %d %d\n",
					       ptr->x + dx - xorigin, 
                                               ptr->y + dy - yorigin, val);
			    else
				fprintf(ofp, "%d %d\n",
                                               ptr->x + dx - xorigin, 
                                               ptr->y + dy - yorigin);
	return;

	case 'R':
          saveformatsign('R', pp->xmin - STARTX, pp->ymin - STARTY, ofp);
	  FOR_CELLS(pp, ptr, dx, dy)
		if (val = getcell(&ptr->cells, dx, dy))
		    if (maxstates > 2)
				fprintf(ofp, "%d %d %d\n",
				       ptr->x + dx - pp->xmin, ptr->y + dy - pp->ymin,
				       val);
		else
		    fprintf(ofp, "%d %d\n",
				       ptr->x + dx - pp->xmin, ptr->y + dy - pp->ymin);
	  return;

	case 'P':
	  savesimple(pp, pp->xmin - STARTX, pp->ymin - STARTY, ofp);
	  return;

	case 'D':
          saveformatsign('D', pp->xmin - STARTX, pp->ymin - STARTY, ofp);
	  x = pp->xmin;
	  y = pp->ymin;
	  FOR_CELLS(pp, ptr, dx, dy)
			if (val = getcell(&ptr->cells, dx, dy)) {
			    if (maxstates > 2)
				fprintf(ofp, "%d %d %d\n",
					       ptr->x + dx - x, ptr->y + dy - y,
					       val);
			    else
				fprintf(ofp, "%d %d\n",
					       ptr->x + dx - x, ptr->y + dy - y);
			    x = ptr->x + dx;
			    y = ptr->y + dy;
			}
	  return;

	case 'M':
          saveformatsign('M', pp->xmin - STARTX, pp->ymin - STARTY, ofp);
	  dy = 0;
	  for (y = pp->ymin; y <= pp->ymax; y++) {
	    for (x = pp->xmin; x <= pp->xmax; x = dx) {
		val = lookcell(pp, x, y);
		for (dx = x + 1; dx <= pp->xmax && lookcell(pp, dx, y) == val; dx++);
		if (val) {
		    if (dy > 1)
			sprintf(p, "%d$", dy);
		    else if (dy > 0)
			strcpy(p, "$");
                    rleflush(70);
		    dy = 0;
		    if (maxstates > 2)
			if (dx - x > 1)
			    sprintf(p, "%d%s", dx - x, rlestr(val));
			else
			    sprintf(p, "%s", rlestr(val));
		    else
			if (dx - x > 1)
			    sprintf(p, "%do", dx - x);
			else
			    strcpy(p, "o");
		}
		else if (dx <= pp->xmax) {
		    if (dy > 1)
			sprintf(p, "%d$", dy);
		    else if (dy > 0)
			strcpy(p, "$");
                    rleflush(70);
		    dy = 0;
		    if (dx - x > 1)
		        sprintf(p, "%db", dx - x);
		    else
		        strcpy(p, "b");
		}
                rleflush(70);
	    }
	    dy++;
	  }
	  if (buf[0])
	    fprintf(ofp, "%s", buf);
	  fprintf(ofp, "!\n");
	  return;

	case 'I':
	  savestructured(pp, ofp);
	  return;

	case 'S':
	  savecomplex(pp, ofp);
	  return;
    }
}

int set_topology(char *ptr) {
       char s1[20], s2[20], t, *p = ptr, c;
       int x, y, dx = 0, dy = 0;

       clr_tile_links();
       if (*p == 'Q')
l1:
          return topology = 'Q', set_torus(0, 0);
       if (*p == 'P' || *p == 'T') {
          t = *p++;
          if (*p == 0) return topology = t, 0;
       }
       else if ((t = topology) == 'Q')
          return 12;
       if (strlen(p) < 2) return 8;
       if (sscanf(p, "%[0-9+-] %*[,*xX] %[0-9+-]", s1, s2) == 2) {
          if (sscanf(s1, "%d %[+-] %d", &x, &c, &dx) == 3) {
             if (c == '-') dx *= -1;
          }
          else if (sscanf(s1, "%d", &x) != 1)
             return 5;
          if (sscanf(s2, "%d %[+-] %d", &y, &c, &dy) == 3) {
             if (c == '-') dy *= -1;
          }
          else if (sscanf(s2, "%d", &y) != 1) return 6;
       }
       else if (sscanf(p, "%*[,*xX] %[0-9+-]", s1) == 1) {
          if (sscanf(s1, "%d %[+-] %d", &y, &c, &dy) == 3) {
             if (c == '-') dy *= -1;
          }
          else if (sscanf(s1, "%d", &y) != 1)
             return 5;
          x = x_max_limit - x_min_limit;
       }
       else if (sscanf(p, "%[0-9+-] %*[,*xX]", s1) == 1) {
          if (sscanf(s1, "%d %[+-] %d", &x, &c, &dx) == 3) {
             if (c == '-') dx *= -1;
          }
          else if (sscanf(s1, "%d", &x) != 1)
             return 4;
          y = y_max_limit - y_min_limit;
       }
       else
          return 11;
       if (x == 0 && y == 0)
          goto l1;
       if (set_torus(x, y)) {
          switch (t) {
          case 'P':
            if (dx != 0 || dy != 0) return 7;
            break;
          case 'T':
            if (dx != 0 || dy != 0) {
               strcpy(inpbuf, "Torus side shift is not supported");
               announce_and_wait(RED_EVER);
               return 8;
            }
          }
          topology = t;
          if (x%8 != 0 || y%8 != 0) {
               sprintf(inpbuf, "Bounded grid size will be rounded to %dx%d",
                       x_max_limit - x_min_limit, y_max_limit - y_min_limit);
               announce_and_wait(RED_EVER);
          }
       }
       return 0;
}
/* tile.c ends here */
